# MIT License
#
# Copyright (c) 2018 Selçuk Karakayalı
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
import sys
from datetime import timedelta


class Progress(object):
    def __init__(self, *args, **kwargs):
        index = kwargs.get('index')
        self.index = index if index else 0
        mxm = kwargs.get('mxm')
        self.max = mxm if mxm else 100
        self.start = time.time()
        self.avg = 0
        self.step_start = self.start
        self.observers = []

    @property
    def elapsed(self):
        return int(time.time() - self.start)

    @property
    def elapsed_td(self):
        return timedelta(seconds=self.elapsed)

    @property
    def progress(self):
        return min(1, self.index / self.max)

    @property
    def percent(self):
        return self.progress * 100

    @property
    def remaining(self):
        return max(0, self.max - self.index)

    def finish(self):
        pass

    def next(self, i=1):
        now = time.time()
        dt = now - self.step_start
        self.step_start = now
        self.index = self.index + i
        self.notify()

    def iter(self, it):
        try:
            for i in it:
                yield i
                self.next()
        finally:
                self.finish()

    def attach(self, observer):
        self.observers.append(observer)

    def detach(self, observer):
        self.observers.remove(observer)

    def notify(self):
        for o in self.observers:
            o.update(self)


class Bar(object):
    fill = '*'
    empty = '-'
    width = 40

    def __init__(self, message=''):
        self.message = message

    def update(self, progress):
        p = int(progress.progress * self.width)
        r = self.width - p

        out = ('\r{}[{}{}]'.format(self.message, p * self.fill,
                                   r * self.empty))
        sys.stdout.write(out)
        sys.stdout.flush()



# MIT License
#
# Copyright (c) 2018 Selçuk Karakayalı
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import time
import base64
import binascii
from urllib.parse import parse_qs, urlunsplit, urlencode
from otp import qr_code
from otp.token import Token, TokenType, InvalidTokenUriError
from otp.progress import Progress, Bar


def get_otp_by_uri(uri):
    if uri.scheme != 'otpauth':
        raise InvalidTokenUriError()

    params = parse_qs(uri.query)

    secret = params.get('secret')
    if secret:
        secret = secret[0]
        del params['secret']
    else:
        raise InvalidTokenUriError('Secret cannot be null')

    kwargs = {}
    kwargs['type'] = uri.netloc
    label = uri.path[1:]
    issuer, user = label.split(':')
    kwargs['issuer'] = issuer
    kwargs['user'] = user

    for k, v in params.items():
        kwargs[k] = v[0]

    return get_otp_by_secret(secret, **kwargs)


def get_otp_by_secret(secret, **kwargs):
    if not secret:
        raise InvalidTokenUriError('Secret cannot be null')

    try:
        secret = base64.b32decode(secret)
    except binascii.Error as e:
        raise InvalidTokenUriError('Invalid secret', secret) from e

    token_type = TokenType.fromValue(kwargs.get('type', 'hotp'))

    issuer = kwargs.get('issuer')
    user = kwargs.get('user')

    period = kwargs.get('period')
    try:
        period = 30 if not period else int(period)
    except ValueError as e:
        raise InvalidTokenUriError('Invalid period', period) from e

    digits = kwargs.get('digits')
    try:
        digits = 6 if not digits else int(digits)
        if digits not in (6, 7, 8):
            raise ValueError
    except ValueError as e:
        raise InvalidTokenUriError('Invalid digits', digits) from e

    algorithm = kwargs.get('algorithm')
    try:
        algorithm = 'SHA1' if not algorithm else algorithm
        if algorithm not in ('SHA1', 'SHA256', 'SHA512'):
            raise ValueError
    except ValueError as e:
        raise InvalidTokenUriError('Invalid encryption algorithm',
                                   algorithm) from e

    return Token(token_type, issuer, user, secret, period, algorithm, digits)


def create_uri(token):
    """
    otpauh://TYPE/LABEL?PARAMS
    where LABEL of form issuer:account
    """
    query = {'secret': base64.b32encode(token.secret), 'period': token.period,
             'algorithm': token.algorithm, 'digits': token.digits}

    path = [e for e in (token.issuer, token.user) if e]

    label = ':'.join(path)

    return urlunsplit(('otpauth', token.type.__str__(),
                       label,
                       urlencode(query), None))


def create_qrcode(token):
    uri = create_uri(token)
    return qr_code.create_qr_code(uri)


def progress(token):
    token_code = token.generateCode()
    current = token_code.progress
    progress = Progress(index=(current * token.period), mxm=token.period)
    message = f'Code = {token_code.code}, progress '
    bar = Bar(message)
    progress.attach(bar)

    # iterate the remaining range
    # let progress bar render until token expires completely
    rem_rng = range(token_code.remaining)
    for i in progress.iter(rem_rng):
        yield i
        time.sleep(1)



# MIT License
#
# Copyright (c) 2018 Selçuk Karakayalı
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
import time
import hmac
import hashlib
import functools
import base64
import math
from enum import Enum


logger = logging.getLogger(__name__)


class TokenCode:
    def __init__(self, code, start, end):
        self.start = start
        self.end = end
        self.code = code
        self.window = end - start

    @property
    def progress(self):
        """
        Returns what progress was made until the expiry of the token
        , between 0 and 1
        """
        current = time.time()

        if current >= self.end:
            return 1

        p = (current - self.start)
        return (p / self.window)

    @property
    def remaining(self):
        """
        Returns the duration in seconds in that token is still valid
        """
        return math.ceil((1 - self.progress) * self.window)


class TokenType(Enum):
        HOTP, TOTP = range(2)

        @staticmethod
        def fromValue(value):
            if value == 'hotp':
                return TokenType.HOTP
            elif value == 'totp':
                return TokenType.TOTP
            else:
                return None

        def __str__(self):
            return self.name.lower()


class Token:
    def __init__(self, t_type, issuer, user, secret,
                 period, algorithm, digits):
        self.type = t_type
        self.issuer = issuer
        self.user = user
        self.secret = secret
        self.period = period
        self.algorithm = algorithm
        self.digits = digits

    def toUri(self):
        """
        otpauh://TYPE/LABEL?PARAMS where
        LABEL is issuer:user
        """
        from urllib.parse import urlunsplit, urlencode
        query = {'secret': base64.b32encode(self.secret), 'period': self.period,
                 'algorithm': self.algorithm, 'digits': self.digits}

        return urlunsplit(('otpauth', self.type.__str__(),
                           self.issuer + ':' + self.user,
                           urlencode(query), None))

    def generateCode(self):
        start = time.time()
        counter = int(start / self.period)
        logger.debug('counter=%s', counter)
        mac = hmac.new(self.secret, counter.to_bytes(8, 'big'), hashlib.sha1)
        digest = mac.digest()
        logger.debug("%s:digest=%s", counter, " ".join([f"{b:02X}" for b in digest]))
        offset = digest[len(digest) - 1] & 0x0F
        logger.debug('%s:offset=%s', counter, offset)

        # mask MSB
        msb = digest[offset] & 0x7F
        for i in range(1, 4):
            msb <<= 8
            msb |= digest[offset + i] & 0xFF

        code = msb % (10 ** self.digits)
        code = f"{code:06d}"

        return TokenCode(code, (counter) * self.period,
                        (counter + 1) * self.period)

    @classmethod
    def fromUri(cls, uri):
        from urllib.parse import parse_qs

        scheme = uri.scheme

        if scheme != 'otpauth':
            raise InvalidTokenUriError('Invalid scheme', scheme)

        token_type = TokenType.fromValue(uri.netloc)
        if not token_type:
            raise InvalidTokenUriError()

        label = uri.path
        if not label:
            raise InvalidTokenUriError()

        issuer, user = label[1:].split(':')

        params = parse_qs(uri.query)

        secret = params.get('secret')
        secret = None if not secret else secret[0]
        if not secret:
            raise InvalidTokenUriError()

        period = params.get('period')
        try:
            period = 30 if not period else int(period[0])
        except ValueError:
            raise InvalidTokenUriError()

        digits = params.get('digits')
        try:
            digits = 6 if not digits else int(digits[0])
            if digits not in (6, 7, 8):
                raise ValueError
        except ValueError:
            raise InvalidTokenUriError()

        issuer_q = params.get('issuer')
        try:
            if issuer_q:
                if issuer and issuer_q[0] != issuer:
                    raise ValueError
                else:
                    issuer = issuer_q[0]
        except ValueError:
            raise InvalidTokenUriError()

        algorithm = params.get('algorithm')
        try:
            algorithm = 'SHA1' if not algorithm else algorithm[0]
            if algorithm not in ('SHA1', 'SHA256', 'SHA512'):
                raise ValueError
        except ValueError:
            raise InvalidTokenUriError()

        return Token(token_type, issuer, user, secret,
                     period, algorithm, digits)

    @classmethod
    def fromString(cls, string):
        from urllib.parse import urlparse, unquote
        return Token.fromUri(urlparse(unquote(string)))

    def __str__(self):
        return f'Token[issuer={self.issuer}, user={self.user}]'

class InvalidTokenUriError(Exception):
    def __init__(self, msg=None, value=None):
        self.args = (msg, value)
        self.msg = msg
        self.value = value



# MIT License
#
# Copyright (c) 2018 Selçuk Karakayalı
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import logging
from urllib.parse import urlparse, unquote
from otp import cli
from otp import core
from otp import configure
from otp.token import InvalidTokenUriError


logger = logging.getLogger(__name__)


def run():
    try:
        args = {k: v for k, v in vars(cli.parse_args()).items() if v}

        logging.basicConfig(level=args.pop('loglevel'))

        cmd = args.pop('command')
        logger.debug('Calling command %s', cmd)
        if cmd == 'get':
            config = configure.read()
            uri = urlparse(unquote(config))
            while True:
                token = core.get_otp_by_uri(uri)
                logger.debug('Built a new token %s', token)
                for i in core.progress(token):
                    pass
        elif cmd == 'qrcode':
            config = configure.read()
            uri = urlparse(unquote(config))
            token = core.get_otp_by_uri(uri)
            logger.debug('Built token %s', token)
            logger.debug(core.create_qrcode(token))
        elif cmd == 'config':
            configure.configure()
        else:
            assert cmd, 'Unknown command {}'.format(cmd)
    except InvalidTokenUriError:
        logger.error('Invalid token', exc_info=1)
    except KeyboardInterrupt:
        pass
    except Exception:
        logger.error('Something\'s wrong here', exc_info=1)


if __name__ == '__main__':
    run()



# MIT License
#
# Copyright (c) 2018 Selçuk Karakayalı
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import pyqrcode
from enum import Enum


class RenderTargetType(Enum):
    TERMINAL = ('terminal')
    TEXT     = ('text')
    SVG      = ('svg')

    def __init__(self, renderer):
        self.renderer = renderer

    def render(self, qr):
        func = getattr(qr, self.renderer)
        return func(quiet_zone=1)

    @staticmethod
    def fromValue(value):
        for r in RenderTargetType:
            if(value.upper() == r.name):
                return r
        else:
            return None


def create_qr_code(url, render_target=RenderTargetType.TERMINAL):
    code = pyqrcode.create(url, error='L', version=5, encoding='utf-8')
    return render_target.render(code)



# MIT License
#
# Copyright (c) 2018 Selçuk Karakayalı
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import io
import base64
import binascii
from otp import core


CONFIG_HOME = os.environ["HOME"] + '/.otp/'
CONFIG_PATH = CONFIG_HOME + '.otprc'


def create_dir_if_not_exists():
    exists = os.path.exists(CONFIG_HOME)
    if not exists:
        os.mkdir(CONFIG_HOME, 0o774)
    return exists


def configure():
    while True:
        secret = input('Secret key: ').strip()
        try:
            base64.b32decode(secret)
            break
        except binascii.Error:
            pass

    issuer = input('Issuer: ').strip()

    user = input('User: ').strip()

    period = input('Period (30 seconds): ').strip()
    if not period:
        period = 30

    digits = input('Digits (6): ').strip()
    if not period:
        digits = 6

    algorithm = input('Algorithm (SHA-256): ').strip()
    if not period:
        algorithm = 'SHA-256'

    create_dir_if_not_exists()

    token = core.get_otp_by_secret(secret, issuer=issuer, user=user,
                                   period=period, digits=digits,
                                   algorithm=algorithm)

    uri = core.create_uri(token)

    abs_path = CONFIG_HOME + '.otprc'
    with io.open(abs_path, 'wt', encoding='UTF-8', newline=None) as f:
        f.write(uri + os.linesep)


def read():
    with io.open(CONFIG_PATH, 'rt', encoding='UTF-8', newline=None) as f:
        line = f.readline()
        return line




#!/usr/bin/python
# -*- coding: utf-8
from setuptools import setup, find_packages

pkg_vars = {}

with open("otp/_version.py") as fp:
    exec(fp.read(), pkg_vars)

setup(
    name='otp-py',
    version=pkg_vars['__version__'],
    author='Selçuk Karakayalı',
    author_email='skarakayali@gmail.com',
    maintainer='Selçuk Karakayalı',
    url='http://github.com/karakays/otp-py/',
    packages=find_packages(),
    install_requires=['pyqrcode>=1.2.1'],
    python_requires='>=3.6',
    license='MIT',
    keywords=['otp', 'otp-codes', 'otp-generator', 'authenticator'],
    description='Generate one time passwords',
    long_description=open('README.rst').read(),
    scripts=['bin/otp'],
    classifiers=[ "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License"]
)
